#!/usr/bin/env python3
import common

import argparse
import sys
import logging
from gitlab import Gitlab, GitlabGetError


def check_if_integration_branch_is_up_to_date(
    manifest_project,
    integration_base,
    project,
    merge_request,
):

    integration_branch = None
    branch_list = []
    if common.is_gitlab_ci_integration_branch(merge_request.source_branch):
        try:
            integration_branch = manifest_project.branches.get(
                merge_request.source_branch,
                retry_transient_errors=True,
            )
            branch_list.append(merge_request.source_branch)
        except GitlabGetError:
            # Branch not found
            pass
    if integration_branch is None:
        integration_branch_name = common.integration_branch_name(
            project.name, merge_request.source_branch
        )
        branch_list.append(integration_branch)
        try:
            integration_branch = manifest_project.branches.get(
                integration_branch_name,
                retry_transient_errors=True,
            )
        except GitlabGetError:
            sys.exit(
                "ERROR: could not find integration branch in {},"
                "branch names checked: {}\n".format(manifest_project.name, branch_list)
            )

    try:
        integration_base_branch = manifest_project.branches.get(
            integration_base, retry_transient_errors=True
        )
    except GitlabGetError:
        sys.exit("ERROR: could not find integration base branch\n")

    integration_base_id = integration_base_branch.commit["id"]

    # Loop over the commits until the integration_branch head id is found
    return common.is_commit_parent_of_project_commit(
        manifest_project, integration_branch.commit["id"], integration_base_id
    )


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--gitlab-url",
        help="""URL to the GitLab instance""",
        dest="gitlab_url",
        required=True,
    )
    parser.add_argument(
        "--token",
        help="""GitLab REST API private access token""",
        dest="token",
        required=True,
    )
    parser.add_argument(
        "--manifest-project",
        help="""name of the manifest project""",
        dest="manifest_project",
        required=True,
    )
    parser.add_argument(
        "--integration-base",
        help="""manifest branch to branch off from""",
        dest="integration_base",
        required=True,
    )
    parser.add_argument(
        "--project",
        help="""name of the project, as specified in the manifest""",
        dest="project",
        required=True,
    )
    parser.add_argument(
        "--merge-request",
        help="""project merge request IID containing the changes to be integrated""",
        dest="merge_request",
        required=True,
    )
    parser.add_argument(
        "--parent-merge-request",
        help="""parent merge requests link, only used for a hint when the check failes""",
        dest="parent_merge_request",
    )
    parser.add_argument(
        "-v",
        "--verbose",
        action="store_true",
        help="""Increase verbosity.""",
    )

    args, _ = parser.parse_known_args()
    if args.verbose:
        logging.basicConfig(level=logging.DEBUG)

    gitlab = Gitlab(args.gitlab_url, private_token=args.token)

    logging.debug(args)
    manifest_project = common.get_project(gitlab, args.manifest_project)
    project = common.get_project(gitlab, args.project)
    merge_request = common.get_merge_request(project, args.merge_request)
    if merge_request is None:
        sys.exit("ERROR: could not get %s  %s" % (project.name, args.merge_request))

    if check_if_integration_branch_is_up_to_date(
        manifest_project=manifest_project,
        integration_base=args.integration_base,
        project=project,
        merge_request=merge_request,
    ):
        print("Integration branch is up to date.")
    else:
        mr_url = merge_request.web_url + "/pipelines"
        if args.parent_merge_request is not None:
            mr_url = args.parent_merge_request + "/pipelines"

        sys.exit(
            "Integration branch is not up to date. Please re-run the MR pipeline:\n"
            "  1. Open the MR pipelines page:\n"
            "     %s\n"
            "  2. Click 'Run Pipeline'" % (mr_url)
        )


if __name__ == "__main__":
    main()
