#!/usr/bin/env python3
import common

import argparse
import logging
import sys
from gitlab import Gitlab, GitlabDeleteError

# from accept_merge_request import accept_merge_request
from create_merge_request import create_merge_request
from get_merge_requests import get_merge_requests
from update_submodule import update_submodule
from update_submodule import get_submodule_project_path_and_revision

# from update_gitlab_ci import update_gitlab_ci_include
from integrate_into_manifest import update_manifest, update_srcrev

from ruamel.yaml import YAML


def read_keys_from_gitlab_ci_yml(gitlab_ci_yml):

    # Read values from existing file
    yaml = YAML()
    data = yaml.load(gitlab_ci_yml)
    logging.debug("Yaml: %s", data)

    try:
        masterbranch = data["variables"]["MASTER_BRANCH_PROJECT"]
        logging.debug("Masterbranch %s", masterbranch)
    except KeyError:
        masterbranch = None
    try:
        recipe = data["variables"]["BB_RECIPE_NAME"]
        logging.debug("Recipe %s", recipe)
    except KeyError:
        recipe = None
    return {"recipe": recipe, "masterbranch": masterbranch}


def update_gitlab_ci_include(content, include_project, new_revision):

    # Remove the SECO-Northern-Europe part from the project filter
    # as it is normally specified by $CI_PROJECT_ROOT_NAMESPACE
    include_project = include_project.split("/", 1)[1]

    yaml = YAML()
    data = yaml.load(content)
    logging.debug("Yaml: %s", data)
    try:
        includes = data["include"]
    except KeyError:
        logging.debug("No include statement found")
        return None
    current_revision = None
    for entry in includes:
        try:
            if include_project in entry["project"]:
                current_revision = entry["ref"]
                break
        except KeyError:
            logging.debug("Failed to parse include statement")
            return None
    if current_revision is None:
        logging.debug("Failed to find %s in include statement", include_project)
        return None

    # Use plain replacement to keep the content of the file
    # Yes, this may fail if the 'current_revision' is used multiple
    # time is this fail. But probably this will not ever happen
    logging.debug("Replace %s with %s", current_revision, new_revision)
    return content.replace(current_revision, new_revision)


def create_gitlab_ci_yml(repo, gitlab_ci_yml):
    """This code snippet was used to initially populate
    all repos with the minial .gitlab-ci.yml that
    includes the files from this repo.
    Currently it is not used and only the revision is
    changed.
    It has also never been test as this function.
    """
    logging.debug("Origin: %s", repo.remotes.origin.url)

    # Contains the base file to be used in the subprojects
    if "ci-test" in repo.remotes.origin.url:
        if "manifest" in repo.remotes.origin.url:
            include_file = "foobar-manifest.yml"
        else:
            include_file = "foobar-manifest-integration.yml"
    else:
        if "manifest" in repo.remotes.origin.url:
            include_file = "manifest.yml"
        else:
            include_file = "manifest-integration.yml"

    # -----------------------------------
    # Adapt content in the include file
    # -----------------------------------
    # Read values from existing file
    yaml = YAML()
    with open(gitlab_ci_yml, "r", encoding="utf8") as fp:
        data = yaml.load(fp)

    logging.debug("Yaml: %s", data)

    try:
        masterbranch = data["variables"]["MASTER_BRANCH_PROJECT"]
        logging.debug("Masterbranch %s", masterbranch)
    except KeyError:
        masterbranch = None
    try:
        recipe = data["variables"]["BB_RECIPE_NAME"]
        logging.debug("Recipe %s", recipe)
    except KeyError:
        recipe = None

    # -----------------------------------
    with open(gitlab_ci_yml, "w", encoding="utf8") as fp:
        fp.write(
            """---
# ---------------------------------------------------------------------------------------
# Include the default CI steps from the gitlab-ci repo
# ---------------------------------------------------------------------------------------
include:
  - project: '${CI_PROJECT_ROOT_NAMESPACE}/yocto/infrastructure/gitlab-ci'
    ref: 49cc4204323bca76190e5ffae1a7d5627157c073
"""
        )
        fp.write("    file: '{}'\n".format(include_file))
        if masterbranch is not None or recipe is not None:
            fp.write("\nvariables:\n")
            if masterbranch is not None:
                fp.write("  MASTER_BRANCH_PROJECT: {}\n".format(masterbranch))
            if recipe is not None:
                fp.write("  BB_RECIPE_NAME: {}\n".format(recipe))

    # -----------------------------------
    repo.git.add(gitlab_ci_yml)

    # ======================================


def deploy_into(project, submodule, revision, branch, create_mr=False):
    """Update the submodule and include refs to the submodule in the given project.
        Create mergerequest if needed.

    Parameters:
        project ( gitlab project): The project which's submodule should be updated
        submodule_name (string): The name of the submodule to pull
        submodule_revision (hex string): The sha hash of the commit to update the submodule to
        branch (string): branch to update, if None, the projects default branch is used

    Returns: tuple of:
        branch (string): Name of the newly created integration branch
        merge_request (gitlab mr): Mergerequest for the integration branch
    """

    # Rewrite update_submodule without git access
    gitlab = project.manager.gitlab
    if branch is None:
        branch = project.default_branch
    logging.debug("Branch: %s", branch)

    (
        submodule_project_path,
        submodule_current_rev,
    ) = get_submodule_project_path_and_revision(project, submodule, branch)

    # Get submodule project
    submodule_project = common.get_project(gitlab, submodule_project_path)

    # Check if revisions are different
    if submodule_current_rev == revision:
        print("Submodule is already at %s" % revision)
        return (None, None, None)

    # Get commits between current and new revision
    revision_range = submodule_current_rev + ".." + revision
    commits = submodule_project.commits.list(
        ref_name=revision_range, retry_transient_errors=True
    )
    if not commits:
        sys.exit("ERROR: no commits found in range %s" % revision_range)
    logging.debug("New commits: %s", commits)

    # Find out if top commit is part of a merge request
    # If so, use source branch of this MR as integration branch name
    # Else use commit sha instead
    integration_branch_suffix = revision
    for mr in commits[0].merge_requests():
        if mr["target_branch"] == submodule_project.default_branch:
            integration_branch_suffix = mr["source_branch"]
            break
    logging.debug("Integration branch suffix: %s", integration_branch_suffix)

    # Check if integration branch already exists and if it is up to date
    integration_branch_name = common.integration_branch_name(
        submodule_project.name, integration_branch_suffix
    )

    integration_branch = None
    for b in project.branches.list(retry_transient_errors=True):
        if b.name == integration_branch_name:
            integration_branch = b
            break

    logging.debug("Existing integration_branch: %s", integration_branch_name)

    if integration_branch:
        print("Replacing integration branch %s" % integration_branch_name)

        try:
            project.branches.delete(integration_branch.get_id())
        except GitlabDeleteError:
            print(
                "Failed to delete branch %s, possible you'll find additional commits in the branch.",
                integration_branch_name,
            )
    else:
        print("Creating integration branch %s" % integration_branch_name)

    logging.debug(
        "Creating integration branch %s from %s",
        integration_branch_name,
        branch,
    )
    project.branches.create({"branch": integration_branch_name, "ref": branch})

    # Update the submodule
    # Write the new revision to the submodule file
    # Construct commit message and commit the change
    message = "Integrate %s/%s%s\n%s" % (
        submodule_project.name,
        integration_branch_suffix,
        " and %d more" % (len(commits) - 1) if len(commits) > 1 else "",
        common.list_commits(commits),
    )

    # ".gitlab-ci",
    project.update_submodule(
        submodule=".gitlab-ci", 
        branch=integration_branch_name,
        commit_sha=revision,
        commit_message=message,
        retry_transient_errors=True,
    )
    #     patch = {
    #         "branch": integration_branch_name,
    #         "commit_message": message,
    #         "actions": [
    #             {
    #                 "action": "update",
    #                 "file_path": submodule,
    #                 "content": "1"
    #                 # revision
    #                 #"bbc15f3863fbcfaab266360e45b54a20c8733bdd"
    #                 #revision,
    #             }
    #         ],
    #     }

    # Now also update the project '.gitlab-ci.yml' file
    gitlab_ci_yml = common.get_repository_file_raw(
        project, ".gitlab-ci.yml", ref=integration_branch_name
    )
    logging.debug(gitlab_ci_yml)

    new_gitlab_ci_yml = update_gitlab_ci_include(
        gitlab_ci_yml,
        submodule_project.web_url.split("//")[1].split("/", 1)[1],
        revision,
    )
    if new_gitlab_ci_yml is None:
        print("Failed to update the include revision in '.gitlab-ci.yml'")
    else:
        logging.debug(new_gitlab_ci_yml)
        # Add the change to the commit list
        #         patch = {
        #             "branch": integration_branch_name,
        #             "commit_message": message,
        #             "actions": [
        #             ],
        #         }
        patch["actions"].append(
            {
                "action": "update",
                "file_path": ".gitlab-ci.yml",
                "content": new_gitlab_ci_yml,
            }
        )

    logging.debug("Send patch: %s", patch["actions"])
    commit = project.commits.create(patch, retry_transient_errors=True)
    logging.debug("Create new commit: %s", commit)
    print()
    sys.exit("So long and thanks for all the fish")
    # Update submodule
    integration_branch, integration_commit, submodule_project = update_submodule(
        project,
        submodule,
        revision,
        branch,
        pre_commit_hook=update_rev_in_gitlab_ci,
    )

    logging.debug("Integration branch: %s, %s", integration_branch, integration_commit)

    # If submodule is already at specified revision, return directly
    if not integration_branch or not create_mr:
        # No commit needed as the submodule is already up to date (integration_branch == None)
        # Or we are done, because no merge request is needed yet
        return integration_branch, integration_commit, None

    # Get source merge request
    mrs = get_merge_requests(
        submodule_project,
        # TODO should this be submodule_project's default branch?
        target_branch="master",
        commit=revision,
    )
    if not mrs:
        sys.exit(
            "ERROR: could not determine source merge request for commit %s" % revision
        )
    source_mr = mrs[0]

    # Create merge request
    # This should be optional
    mr, created = create_merge_request(
        project, integration_branch, project.default_branch
    )
    if created:
        common.crosslink_merge_requests(source_mr, mr)
        print("Created new merge request:\n%s" % mr.web_url)
    else:
        print("Existing integration merge request:\n%s" % mr.web_url)
    return integration_branch, integration_commit, mr


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--gitlab-url",
        help="""URL to the GitLab instance""",
        dest="gitlab_url",
        required=True,
    )
    parser.add_argument(
        "--token",
        help="""GitLab REST API private access token""",
        dest="token",
        required=True,
    )
    parser.add_argument(
        "--project",
        "--manifest-project",
        help="""name of the GitLab project""",
        dest="project",
    )
    parser.add_argument(
        "--submodule",
        help="""submodule to update""",
        dest="submodule",
        required=True,
    )
    parser.add_argument(
        "--revision",
        help="""new revision for submodule""",
        dest="revision",
        required=True,
    )
    parser.add_argument(
        "--branch",
        help="""project branch (if not default branch)""",
        dest="branch",
        required=False,
        default=None,
    )
    parser.add_argument(
        "--merge",
        help="""if set, perform merge after integration""",
        dest="merge",
        action="store_true",
        required=False,
        default=False,
    )
    parser.add_argument(
        "--manifest-file",
        help="""manifest file name (default: 'default.xml')""",
        dest="manifest_file",
        default=common.manifest_file,
        required=False,
    )
    parser.add_argument(
        "--srcrev-file",
        help="""source revision file name (default: 'SRCREV.conf')""",
        dest="srcrev_file",
        default=common.srcrev_file,
        required=False,
    )
    parser.add_argument(
        "projects",
        help="""List of projects the change should be deployed to additionally
                to the manifest project given as named parameter.""",
        nargs="*",
    )
    parser.add_argument(
        "-v",
        "--verbose",
        action="store_true",
        help="""Increase verbosity.""",
    )

    args, _ = parser.parse_known_args()
    if args.verbose:
        logging.basicConfig(level=logging.DEBUG)

    gitlab = Gitlab(args.gitlab_url, private_token=args.token)

    project_integration = {}
    # Update submodule in this project, create MR if needed TODO
    for p in args.projects + [args.project]:
        logging.debug("Integrate into: %s", p)

        gitlab_project = common.get_project(gitlab, p)

        integration_branch, integration_commit, mr = deploy_into(
            gitlab_project,
            args.submodule,
            args.revision,
            args.branch,
        )
        if integration_branch is not None:
            project_integration[p] = {
                "project": gitlab_project,
                "mr": mr,
                "branch": integration_branch,
                "commit": integration_commit,
            }
            logging.debug(
                "Integration branch: %s (%s)", integration_branch, integration_commit
            )

    # If submodule is already at specified revision in all projects, exit successfully
    if len(project_integration) == 0:
        print("No integration done, changes are already included in all projects.")
        sys.exit(0)

    # Now create the integration commit in the manifest
    # for all subprojects at once
    manifest_project = project_integration[args.project]
    manifest = common.get_repository_file_raw(
        manifest_project["project"], args.manifest_file, ref=manifest_project["branch"]
    )
    logging.debug(manifest)
    srcrev = common.get_repository_file_raw(
        manifest_project["project"], args.srcrev_file, ref=manifest_project["branch"]
    )
    logging.debug(srcrev)

    for p in args.projects:
        integration = project_integration[p]
        logging.debug(
            "Update %s to %s", integration["project"].name, integration["commit"]
        )

        new_manifest = update_manifest(
            manifest, integration["project"], integration["commit"]
        )
        if new_manifest is not None:
            manifest = new_manifest
            logging.debug(manifest)
            continue

        # get BB_RECIPE_NAME from the projects .gitlab-ci.yml
        gitlab_ci_yml = common.get_repository_file_raw(
            integration["project"], ".gitlab-ci.yml", ref=integration["branch"]
        )
        project_keys = read_keys_from_gitlab_ci_yml(gitlab_ci_yml)

        new_srcrev = update_srcrev(
            srcrev, project_keys["recipe"], integration["commit"]
        )
        if new_srcrev is not None:
            srcrev = new_srcrev
            logging.debug(srcrev)
        else:
            logging.debug("Project %s not found in xml or srcrev file", p)

    # Now manifest and srcrev contain the updated revisions created above
    # Send them to gitlab as patch
    patch = {
        "branch": manifest_project["branch"],
        "commit_message": "This is a test",  # TODO
        "actions": [
            {
                "action": "update",
                "file_path": args.manifest_file,
                "content": manifest,
            },
            {
                "action": "update",
                "file_path": args.srcrev_file,
                "content": srcrev,
            },
        ],
    }
    logging.debug("Send patch: %s", patch)
    commit = manifest_project["project"].commits.create(
        patch, retry_transient_errors=True
    )
    logging.debug("Created new commit: %s", commit)

    sys.exit(0)

    if not args.merge:
        print(
            "Skipping automatic merge in MR context. If you like to extend the "
            "integration MR by hand, please do it now. Afterwards you can either merge "
            "it by hand or re-run this job on the master branch after the source MR "
            "has been merged."
        )
        sys.exit(0)

    # The manifest needs to be merged at last
    # TODO this is currently not used
    # merge_requests.append(merge_request_manifest)
    # for mr in merge_requests:
    #    logging.debug("Merge %s", mr)
    #    # Wait until GitLab has checked merge status
    #    common.wait_until_merge_status_is_set(project, mr)

    #    # Attempt to merge
    #    merged = accept_merge_request(project, mr, rebase=True)

    #    if not merged:
    #        sys.exit(
    #            "Integration MR could not be merged. You have two possibilities to fix "
    #            "this:\n"
    #            "  1. Checkout the MR and rebase it on the current master manually, or\n"
    #            "  2. Delete the MR (Edit -> Delete in the MR UI)\n"
    #            "In either case restart this job afterwards in order to get it merged."
    #        )

    print("Successfully merged")


if __name__ == "__main__":
    main()
