#!/usr/bin/env python3
import argparse
import sys
from furl import furl
from lxml import etree


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--manifest",
        help="""manifest file to parse projects from""",
        dest="manifest",
        required=True,
    )
    parser.add_argument(
        "--remote",
        help="""get only projects with this remote""",
        dest="remote",
        required=False,
    )
    parser.add_argument(
        "--concat-namespaces",
        help="""parse namespace from fetch URL and prepend it to project names""",
        dest="concat_namespaces",
        action="store_true",
        required=False,
    )

    args, _ = parser.parse_known_args()

    # Parse manifest file
    try:
        manifest = etree.parse(args.manifest)
    except FileNotFoundError:
        sys.exit("ERROR: file '%s' not found" % args.manifest)

    # Get namespace from remote
    # This is needed for cases where the remote URL contains a part of the project
    # namespace (e.g. "ssh://git@gitlab.com/garz-fricke/yocto") and the project name
    # contains another part of it (e.g. "layers/meta-guf-machine"). There is no GitLab
    # API call which will find a project given this information.
    # Thus we are adding a possibility to parse the namespace from the remote and pre-
    # pend it to the project name in order to return it including its complete namespace
    # (e.g. "garz-fricke/yocto/layers/meta-guf-machine").
    if args.concat_namespaces:
        remote = manifest.find("remote[@name='%s']" % args.remote)
        if remote is None:
            sys.exit("ERROR: remote '%s' not found in manifest" % args.remote)
        path = furl(remote.get("fetch")).path
        prefix = str(path).strip("/") + "/"
    else:
        prefix = ""

    # Find project references in manifest
    if args.remote:
        find_expression = "project[@remote='%s']" % args.remote
    else:
        find_expression = "project"
    projects = manifest.findall(find_expression)

    for project in projects:
        print(prefix + project.get("name"))


if __name__ == "__main__":
    main()
